/*
 This file is part of the digitalSTROM ds485-daemon (ds485d).

 Copyright (C) 2010 digitalSTROM.org, Schlieren, Switzerland

 digitalSTROM ds485-daemon is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation version 2
 of the License.

 digitalSTROM ds485-daemon is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with digitalSTROM ds485-daemon; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "config.h"

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <signal.h>
#include <setjmp.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>
#include <pthread.h>
#include <fcntl.h>
#include <errno.h>
#include <assert.h>
#include <locale.h>

#include <digitalSTROM/dsuid.h>
#include <digitalSTROM/ds485-core.h>
#include <digitalSTROM/ds485-socket-server.h>

#include "base/discovery.h"
#include "base/ds_string.h"
#include "base/log.h"
#include "base/util.h"


/******************************************************************************
 *
 */
static ds485Handle_t busHandle = NULL;

extern char *optarg;
extern int optind;

static int debugLevel = 5;
int mainActive = 1;
int gotSignal = 0;
static jmp_buf restartbuf;

static pthread_t rxThread;

static pthread_t discoveryThread;

/* Serial Device Name of the device we manage */
static char* serialDevice;

/* Some serial device (adapters) cannot receive own transmission data */
static int serialLoopback = 0;

/* process id file */
static char* pidfile = "/var/run/ds485d.pid";


/* listen on all network listen, or just loopback */
static int listenGlobal = 0;

/* configured dsid */
static dsid_t dsid = { {0x35, 0x04, 0x17, 0x5f, 0xe0, 0, 0, 0xf0, 0, 0, 0, 1} };
static dsuid_t dsuid = { {0x35, 0x04, 0x17, 0x5f, 0xe0, 0, 0, 0, 0, 0, 0, 0xf0, 0, 0, 0, 1, 0} };

/* default port for ds485d */
static int default_port = 8442;

/* assigned port */
static int port = 0;

/* flag for avahi service*/
static int avahiEnable = 0;

/**
 * Server data
 */
struct dS485d {
    ds485Handle_t* busHandle;    /* ds485 lib handle */
    struct dS485_server *netlibHandle;
};

static struct dS485d dS485d_data;

/* Name of (unprivileged) user to change to when we drop privileges. */
#ifndef DS485D_USER
#define DS485D_USER    "nobody"
#endif


/********************* SIGNAL & ERROR HANDLING *********************/

/*
 * Get the signal and save for main-loop processing
 */
void onsignal(int signo)
{
    gotSignal = signo;
    mainActive = 0;
    ds485_server_abort(dS485d_data.netlibHandle);
}

static void prepare_dsmapi_response(ds485_container_t *container,
                                    const dsuid_t *destinationId,
                                    uint8_t transactionId,
                                    uint8_t messageId,
                                    uint8_t modifierId)
{
    container->containerType = DS485_CONTAINER_RESPONSE;
    container->destinationId = *destinationId;
    container->sourceId = dsuid;

    container->raw_data[0] = transactionId;
    container->raw_data[1] = messageId;
    container->raw_data[2] = modifierId;
    if (messageId == 160) {
        /* get bus member request */
        container->raw_data[3] = 0; // ERROR_OK
        container->raw_data[4] = 1; // BusMember_ds485d
        container->length = 4;
    } else {
        /* unhandled requests */
        container->raw_data[3] = 0x16; // ERROR_REQUEST_CAN_NOT_BE_EXECUTED
        container->length = 3;
    }
}

static void handle_request_for_ds485d(struct dS485_server *server,
                                      const ds485_container_t *packet)
{
    ds485_container_t container;
    CONTAINER_STR_BUF(buf);
    uint8_t transactionId = packet->raw_data[0];
    uint8_t messageId = packet->raw_data[1];
    uint8_t modifierId = packet->raw_data[2];

    report(LOG_DEBUG, "Netlib message addressed dS485d\n");
    sprintf_container(buf, packet);
    report(LOG_DEBUG, "> container: %s\n", buf);

    prepare_dsmapi_response(&container, &packet->sourceId, transactionId, messageId, modifierId);
    sprintf_container(buf, &container);
    report(LOG_DEBUG, "< container: %s\n", buf);
    /* deliver to each client session */
    ds485_server_reply(server, &container);
}

static void send_response_from_ds485d(ds485Handle_t* busHandle,
                                      const ds485_container_t *packet)
{
    ds485_container_t container;
    CONTAINER_STR_BUF(buf);
    uint8_t transactionId = packet->raw_data[0];
    uint8_t messageId = packet->raw_data[1];
    uint8_t modifierId = packet->raw_data[2];

    report(LOG_DEBUG, "Serial message addressed dS485d\n");
    sprintf_container(buf, packet);
    report(LOG_DEBUG, "> container: %s\n", buf);

    prepare_dsmapi_response(&container, &packet->sourceId, transactionId, messageId, modifierId);
    sprintf_container(buf, &container);
    report(LOG_DEBUG, "< container: %s\n", buf);
    /* deliver to serial interface */
    int error = ds485_api_send_data(busHandle, &container);
    if (error) {
        CONTAINER_STR_BUF(buf);
        if (error == DS485_INTERNAL_ERROR_GENERAL) {
            report(LOG_ERR, "send_response_from_ds485d returned -1. " \
                            "handle %d, length %d\n", busHandle, packet->length);
        } else if (error == DS485_INTERNAL_ERROR_EMPTY_TX_QUEUE) {
            report(LOG_ERR, "send_response_from_ds485d returned -2. " \
                            "tx response queue was empty\n");
        } else if (error == DS485_INTERNAL_ERROR_INVALID_BUS_ADDR) {
            report(LOG_ERR, "send_response_from_ds485d returned -3. " \
                            "invalid bus address\n");
        } else {
            report(LOG_ERR, "send_response_from_ds485d returned %d\n", error);
        }
        sprintf_container(buf, packet);
        report(LOG_ERR, "container: %s\n", buf);
    }
}

static bool is_request_for_ds485d(const ds485_container_t *packet)
{
    if ((memcmp(&dsuid, &packet->destinationId, sizeof(dsuid)) == 0) &&
        (packet->containerType == DS485_CONTAINER_REQUEST)) {
        return true;
    }
    return false;
}

/*************************** DS485 SERVER IMPL ***********************/
int ds485_server_send_data(struct dS485_server *server,
                           const ds485_container_t *packet)
{
    (void)server;
    if (is_request_for_ds485d(packet)) {
        handle_request_for_ds485d(dS485d_data.netlibHandle, packet);
        return 0;
    }
    int error = ds485_api_send_data(dS485d_data.busHandle, packet);
    if (error) {
        CONTAINER_STR_BUF(buf);
        if (error == DS485_INTERNAL_ERROR_GENERAL) {
            report(LOG_ERR, "broadcast send data returned -1. " \
                            "handle %d, length %d\n", dS485d_data.busHandle, packet->length);
        } else if (error == DS485_INTERNAL_ERROR_EMPTY_TX_QUEUE) {
            report(LOG_ERR, "broadcast send data returned -2. " \
                            "tx response queue was empty\n");
        } else if (error == DS485_INTERNAL_ERROR_INVALID_BUS_ADDR) {
            report(LOG_ERR, "broadcast send data returned -3. " \
                            "invalid bus address\n");
        } else {
            report(LOG_ERR, "broadcast send data returned %d\n", error);
        }
        sprintf_container(buf, packet);
        report(LOG_ERR, "container: %s\n", buf);
    }
    return error;
}

int ds485_server_send_long_data(struct dS485_server *server,
                                const ds485n_packet_t* packet)
{
    (void)server;
    (void)packet;
    return 0;
}

int ds485_server_get_bus_members(struct dS485_server *server,
                                 dsuid_t *devices, int length)
{
    (void)server;
    return ds485_api_get_bus_members(dS485d_data.busHandle, devices, length);
}

int ds485_server_restart_bus(struct dS485_server *server)
{
    (void)server;
    report(LOG_NOTICE, "restart serial bus\n");
    return ds485_api_restart_bus(dS485d_data.busHandle);
}


dsuid_t ds485_server_get_dsuid(struct dS485_server *server)
{
    (void)server;
    return dsuid;
}

bus_state_t ds485_server_get_bus_state(struct dS485_server *server)
{
    (void)server;
    return ds485_api_get_bus_state(dS485d_data.busHandle);
}

int ds485_server_set_packetfilter(struct dS485_server *server, uint8_t rxfilter)
{
    /* TODO, is do nothing okay? */
    (void)server;
    (void)rxfilter;
    return 0;
}

struct dS485_server_callbacks cb =
{
    .receive_data = ds485_server_send_data,
    .receive_long_data = ds485_server_send_long_data,
    .get_bus_members = ds485_server_get_bus_members,
    .restart_bus = ds485_server_restart_bus,
    .get_dsuid = ds485_server_get_dsuid,
    .get_bus_state = ds485_server_get_bus_state,
    .set_packetfilter = ds485_server_set_packetfilter,
};

static void print_usage(void)
{
    fprintf(stderr, "Usage: dS485d [OPTION] <device>\n");
    fprintf(stderr, "       <device> is a tty name\n\n");
    fprintf(stderr, "       OPTION is a selection out of the following options:\n");
    fprintf(stderr, "       -h  --  print this help\n");
    fprintf(stderr, "       -d  --  debug-level: max debug level still logged: 0-7 (default: 5)\n");
    fprintf(stderr, "       -p  --  PID-file (default: /var/run/ds485d.pid)\n");
    fprintf(stderr, "       -F  --  stay in foreground (default: not set == daemonize)\n");
    fprintf(stderr, "               log-output: foreground: stdout\n");
    fprintf(stderr, "                           daemonized: syslog\n");
    fprintf(stderr, "       -G  --  enable external access: accept socket connection from other machines\n");
    fprintf(stderr, "       -L  --  enable extra serial loopback on RS485 interface\n");
    fprintf(stderr, "       -V  --  print ds485d version information\n");
    fprintf(stderr, "       -i  --  change dSID of ds485d (default: 3504175fe00000f000000001\n");
    fprintf(stderr, "       -a  --  enable avahi discovery service\n");
    fprintf(stderr, "\n");
}

/*************************** Protocol Callbacks *******************************/
void callback_bus_change(ds485Handle_t dev, dsuid_t *dsuid, int flags)
{
    DSUID_STR(buf);
    (void)dev;

    dsuid_to_string(dsuid, buf);
    report(LOG_NOTICE, "bus change: %s -> %i\n", buf, flags);

    /*
     * TODO http://redmine.digitalstrom.org/issues/2140
     */
    ds485_server_emit_bus_change(dS485d_data.netlibHandle, *dsuid, flags);
}

/**
 * registered system callback for protocol (debug) messages
 */
int callback_system_msg(ds485Handle_t dev, unsigned char level, const char *msg, va_list args)
{
    char buffer[512];
    (void)dev;

    vsnprintf(buffer, 512, msg, args);
    report(level, "%s", buffer);
    return 0;
}

/**
 * registered device callback for bus state changes
 */
int callback_state(ds485Handle_t dev, bus_state_t state)
{
    (void)dev;
    report(LOG_NOTICE, "bus state change: %i\n", state );
    ds485_server_emit_bus_state(dS485d_data.netlibHandle, state);
    return 0;
}

/******************************* RECEIVE THREAD *******************************/

static void filter_packet(ds485_container_t *cont)
{
    unsigned char *cmd = cont->data; /* cmd offset-0; modifier offset-1 */
    unsigned char *d = &cmd[3]; /* data starts at offset 3 */

    /*
     * http://redmine.digitalstrom.org/issues/7630
     * prefix it seems this happens, when version is '0'
     */
    if ((cont->containerType !=  DS485_CONTAINER_RESPONSE) ||
        (cmd[0] != 0x20))  {/* DEVICE_INFO command */
        return;
    }

    if ((d[4] == 0 && d[5] == 0) || /* product id */
        (d[6] == 0 && d[7] == 0) || /* function id */
        (d[8] == 0 && d[9] == 0) || /* version */
        (memcmp(&d[45], DSUID_NULL.id, DSUID_SIZE) == 0)) {
        CONTAINER_STR_BUF(buf);
        char *p;
        int i;

        sprintf_container(buf, cont);
        report(LOG_WARNING, "suspect device info pkt:%s\n", buf);

        p = buf;
        for (i = 0; i < 0x41; i++) {
            p += sprintf(p, "%02x", cmd[i]);
        }
        report(LOG_WARNING, " --   full dsm api data:%s\n", buf);
    }
}

static void* sysRxThread(void* arg)
{
    ds485Handle_t *handle = (ds485Handle_t *) arg;
    ds485_container_t container;
    int rval;

    /* Read incoming ds485 data */
    while (mainActive) {
        rval = ds485_api_recv_data(handle, &container);
        if (rval < 0) {
            report(LOG_NOTICE, "ds485_api_recv_data failed %d\n", rval);
            continue;
        }

        filter_packet(&container);
        assert(container.length <= DS485_MAX_PAYLOAD_SIZE);

        if (is_request_for_ds485d(&container)) {
            send_response_from_ds485d(dS485d_data.busHandle, &container);
        } else {
            /* deliver to each client session */
            ds485_server_reply(dS485d_data.netlibHandle, &container);
        }
    }
    report(LOG_NOTICE, "ds485 rx thread end.\n");
    return NULL;
}

/******************************* DISCOVERY THREAD *****************************/
static void* sysDiscoveryThread(void *arg __attribute__((unused)))
{
    /* Discovery setup */
    ds485d_t discoveryHandle;
    memset(&discoveryHandle, 0, sizeof(struct ds485d));

    DSUID_STR(name_dsid);
    dsuid_to_string(&dsuid, name_dsid);

    discoveryHandle.port = port;
    discoveryHandle.ds485d_dsuid = strdup(name_dsid);

    while (mainActive) {
        if (ds485d_discovery_init(&discoveryHandle, "ds485d") != DS485D_DISCOVERY_OK) {
            sleep(1);
            continue;
        }
        /* ds485d_discovery_work will block until avahi client stops */
        ds485d_discovery_work(&discoveryHandle);
    }

    free(discoveryHandle.ds485d_dsuid);
    discoveryHandle.ds485d_dsuid = NULL;

    ds485d_discovery_cleanup(&discoveryHandle);

    report(LOG_NOTICE, "ds485 avahi thread end.\n");
    return NULL;
}

/******************************* MAIN *******************************/

/*
 * Function main (argc, )
 *
 *    Main function
 *
 */
int main(int argc, char* argv[])
{
    int c, rval;
    int goBackground = 1;
    static int st;

    init_logger("ds485d", LOG_INFO, false);

    mainActive = 1;

#ifdef HAVE_SETLOCALE
    setlocale(LC_NUMERIC, "C");
#endif

    if (argc < 2) {
        print_usage();
        exit(-1);
    }

    /* Look for options */
    while ((c = getopt(argc, argv, "ahd:p:FGLVi:")) != -1) {
        switch (c) {
        case 'h':
            print_usage();
            exit(0);
        case 'd':
            debugLevel = atoi(optarg);
            break;
        case 'p':
            pidfile = strdup(optarg);
            break;
        case 'F':
            goBackground = 0;
            break;
        case 'G':
            listenGlobal = 1;
            break;
        case 'L':
            serialLoopback = 1;
            break;
        case 'i':
        {
            rval = scanf_dsid(optarg, &dsid);
            if (rval < 0) {
                report(LOG_ERR, "dSID in the wrong format\n");
                exit(-1);
            }
            dsuid = ds485_dsid_to_dsuid(&dsid);
            break;
        }
        case 'a':
        {
            avahiEnable = 1;
            break;
        }
        case 'V':
            printf("Version: v%s\n", VERSION);
            exit(0);
        default:
            exit(-1);
        }
    }

    set_log_level(debugLevel);

    /* Last arg is device or file name. */
    serialDevice = strdup(argv[argc-1]);

    /* go to background mode */
    if (goBackground) {
        enable_syslog(true);
        daemonize();
    }
    report(LOG_NOTICE, "%s, %s (built %s, %s)\n", PACKAGE_NAME, VERSION, __DATE__, __TIME__);

    /* PID file */
    if (pidfile) {
        FILE* fp;

        if ((fp = fopen(pidfile, "w")) != NULL) {
            fprintf(fp, "%u\n", (unsigned int) getpid());
            fclose(fp);
        } else {
            report(LOG_ERR, "Cannot create PID file: %s\n", pidfile);
        }
    }

    /* got hangup signal */
    if ((st = setjmp(restartbuf)) > 0) {
        report(LOG_WARNING, "ds485d restarted by SIGHUP\n");
    }

    /* Trap signals */
    gotSignal = 0;
    signal(SIGINT, onsignal);
    signal(SIGTERM, onsignal);
    signal(SIGQUIT, onsignal);
    signal(SIGHUP, onsignal);
    signal(SIGPIPE, SIG_IGN);

    /*
     * ensure we have a valid handle before subscribing
     * to ds485-core callbacks
     */
    dS485d_data.netlibHandle = ds485_server_init(&cb, false);
    if (!dS485d_data.netlibHandle) {
        report(LOG_ERR, "Failed to initialize netlib server\n");
        exit(-4);
    }

    /* start ds485 protocol */
    busHandle = ds485_api_open(serialDevice, &dsid);
    if (NULL == busHandle) {
        report(LOG_ERR, "ERROR initializing dS485 protocol\n");
        exit(2);
    }
    dS485d_data.busHandle = busHandle;

    /* register message handler */
    ds485_api_ioctl(busHandle, IOC_SET_MSG_CB,
                    callback_system_msg, sizeof(callback_system_msg));

    ds485_api_ioctl(busHandle, IOC_SET_BUS_CHANGE_CB,
                    callback_bus_change, sizeof(callback_bus_change));

    ds485_api_ioctl(busHandle, IOC_SET_BUS_STATE_CB,
                    callback_state, sizeof(callback_state));

    /* configure software loopback */
    if (serialLoopback) {
        ds485_api_ioctl(busHandle, IOC_LOOPBACK_MODE,
                        &serialLoopback, sizeof(serialLoopback));
    }

    /* Start receiver thread */
    if (pthread_create(&rxThread, NULL, sysRxThread, busHandle)) {
        report(LOG_ERR, "Could not create serial thread.\n");
        mainActive = 0;
        goto out;
    }
    pthread_detach(rxThread);

    /* TODO sscanf_port */
    port = default_port;

    if (avahiEnable)
    {
        /* Start discovery thread */
        if (pthread_create(&discoveryThread, NULL, sysDiscoveryThread, NULL))
        {
            report(LOG_ERR, "Could not create discovery thread.\n");
            mainActive = 0;
            goto out;
        }
        pthread_detach(discoveryThread);
    }

    if ((rval = ds485_server_listen(dS485d_data.netlibHandle, port, listenGlobal))) {
        report(LOG_ERR, "Error listen for clients failed %d\n", rval);
        mainActive = 0;
        exit(2);
    }

out:
    switch (gotSignal) {
    case SIGTERM:
    case SIGINT:
        report(LOG_NOTICE, "got SIGTERM or SIGINT\n");
        mainActive = 0;
        break;
    case SIGHUP:
        report(LOG_NOTICE, "got SIGHUP\n");
        longjmp(restartbuf, 1);
        /* never reached */
        break;
    default:
        break;
    }

    ds485_server_abort(dS485d_data.netlibHandle);

    /* Cleanup threads */
    pthread_cancel(rxThread);

    if (avahiEnable)
    {
        pthread_cancel(discoveryThread);
    }

    /* shutdown ds485 protocol */
    ds485_api_close(busHandle);

    ds485_server_cleanup(dS485d_data.netlibHandle);
    cleanup_logger();

    return 0;
}
